# -*- coding: utf-8 -*-
__app_name__ = 'pylinkam'
__version__ = '1.2.6'

__author__ = 'Chris Harrison'
__credits__ = [__author__]

__maintainer__ = 'Chris Harrison'
__email__ = '629204+ravngr@users.noreply.github.com'

__status__ = 'Development'
